from argparse import (
    ArgumentDefaultsHelpFormatter,
    ArgumentParser,
    RawTextHelpFormatter,
)
from importlib.metadata import version
from textwrap import dedent

from brag.defaults import Defaults

__version__ = version("pybrag")


class BragCliFormatter(RawTextHelpFormatter, ArgumentDefaultsHelpFormatter):
    pass


def parse_args():
    defaults = Defaults()

    parser = ArgumentParser("brag")
    parser.add_argument("-v", "--version", action="store_true")

    def add_ssl_verify(subparser):
        subparser.add_argument(
            "--no-ssl-verify",
            action="store_true",
            help="if present, litellm.ssl_verify will be set to False.",
        )

    def add_llm(subparser):
        subparser.add_argument(
            "--llm",
            help=dedent(
                """\
                LLM model in litellm format (provider/model_id). e.g.,
                * openai/gpt-40-mini
                * ollama/llama3.1:8b
                * hosted_vllm/meta-llama/Llama-3.1-8B-Instruct
                """
            ),
            default="openai/gpt-4o-mini",
        )

    def add_temperature(subparser):
        # Not setting default as some openai reasoning (o1, o3, etc.) models do
        # not accept temperature.
        subparser.add_argument(
            "--temperature",
            type=float,
            help="LLM temperature.",
        )

    def add_api_key(subparser):
        subparser.add_argument(
            "--api-key",
            help=dedent(
                """\
                LLM API Key (e.g. OPENAI_API_KEY). These environment variables
                are recognized so can be omitted here:

                * OPENAI_API_KEY

                If using ollama/vllm, an api key may be required.
                """
            ),
        )

    def add_base_url(subparser):
        subparser.add_argument(
            "--base-url",
            help=dedent(
                """\
                Base url (e.g. openai base url).

                For ollama, brag automatically sets to: http://localhost:11434
                """
            ),
        )

    def add_llm_arg_parsers(subparser):
        add_llm(subparser)
        add_temperature(subparser)
        add_base_url(subparser)
        add_api_key(subparser)

    def add_emb(subparser):
        subparser.add_argument(
            "--emb",
            type=str,
            help=dedent(
                """\
                Embedding model in litellm format (provider/model_id). e.g.,
                * openai/text-embedding-3-small
                * ollama/nomic-embed-text
                * hosted_vllm/BAAI/bge-m3
                """
            ),
            default="openai/text-embedding-3-small",
        )

    def add_emb_base_url(subparser):
        subparser.add_argument(
            "--emb-base-url",
            help=dedent(
                """\
                Base url for embedder. For example, 
                * http://localhost:8001
                * http://some-random-site/

                You will have to repeat the url even if it is the same as
                --base-url.
                """
            ),
        )

    def add_emb_api_key(subparser):
        subparser.add_argument(
            "--emb-api-key",
            help=dedent(
                """\
                API key for embedder. You will have to repeat the api key even if
                it is the same as --api-key.

                These environment variables are recognized so can be omitted
                here:

                * OPENAI_API_KEY
                """
            ),
        )

    def add_emb_arg_parsers(subparser):
        add_emb(subparser)
        add_emb_api_key(subparser)
        add_emb_base_url(subparser)

    def add_log(subparser):
        subparser.add_argument("--log", help="logging level", default="WARNING")

    def add_chunksize(subparser):
        subparser.add_argument(
            "--chunk-size",
            help="chunk size for document indexing.",
            type=int,
            default=defaults.chunk_size,
        )

    def add_chunk_overlap(subparser):
        subparser.add_argument(
            "--chunk-overlap",
            help="chunk-overlap for indexing.",
            type=int,
            default=defaults.chunk_overlap,
        )

    def add_batchsize(subparser):
        subparser.add_argument(
            "--batchsize",
            help="how many document chunks to index at a time.",
            default=defaults.batch_size,
            type=int,
        )

    def add_corpus_dir(subparser):
        subparser.add_argument(
            "--corpus-dir", help="directory to corpus.", default="."
        )

    def add_db_dir(subparser):
        subparser.add_argument(
            "--db-dir",
            help="directory to save vector database.",
            default=".brag/db",
        )

    def add_db_name(subparser):
        subparser.add_argument(
            "--db-name",
            help="vector database collection name.",
            type=str,
        )

    def add_db_host(subparser):
        subparser.add_argument(
            "--db-host",
            help=dedent("""\
                        Hostname for chromadb server. If not provided and
                        --db-port is provided, this will default to 'localhost'.
                        If --db-port is not provided, this flag is ignored. 
                        """),
            type=str,
        )

    def add_db_port(subparser):
        subparser.add_argument(
            "--db-port",
            help=dedent("""\
                        Port for chromadb server. If not supplied, chromadb will
                        run locally.
                        """),
            type=int,
        )

    def add_num_retrieved_docs(subparser):
        subparser.add_argument(
            "-n",
            "--num-retrieved-docs",
            help=dedent(
                """\
                number of document chunks to retrieve. Recommendation: Use an
                LLM with a context window of at least 65K (e.g. the llama3.1,
                llama3.2, llama3.3.).  Number of retrieved documents should
                result in approx 5000-6000 chunks.  e.g. if using a chunksize of
                340, using about 15 retrieved docs would result in ~5100 chunks.
                batchsize * chunksize should be about 50000 chunks.  So if using
                a chunksize of 340, use a batch size of about 147.
                """
            ),
            type=int,
            default=defaults.num_retrieved_docs,
        )

    def add_no_cache_db(subparser):
        subparser.add_argument(
            "--no-cache-db",
            action="store_true",
            help="if present, vector database is stored in memory and is NOT saved.",
        )

    # Sub commands.
    parsers = parser.add_subparsers(dest="command", required=False)

    # Delete index.
    rm_index_parser = parsers.add_parser(
        "rm-index",
        help="Delete a ChromaDB collection.",
        formatter_class=BragCliFormatter,
    )
    add_db_dir(rm_index_parser)
    add_db_name(rm_index_parser)
    add_db_host(rm_index_parser)
    add_db_port(rm_index_parser)
    add_log(rm_index_parser)
    rm_index_parser.add_argument(
        "--reset",
        action="store_true",
        help="If present, all collections in vector database are removed.",
    )

    # Index a directory of documents.
    index_parser = parsers.add_parser(
        "index",
        help="index a directory of docs.",
        formatter_class=BragCliFormatter,
    )
    add_api_key(index_parser)
    add_base_url(index_parser)
    add_emb_arg_parsers(index_parser)
    add_corpus_dir(index_parser)
    add_db_dir(index_parser)
    add_db_name(index_parser)
    add_db_host(index_parser)
    add_db_port(index_parser)
    add_batchsize(index_parser)
    add_chunksize(index_parser)
    add_chunk_overlap(index_parser)
    add_log(index_parser)
    add_ssl_verify(index_parser)
    index_parser.add_argument(
        "-o",
        "--overwrite",
        action="store_true",
        help="whether or not to overwrite existing vector database.",
    )

    # Ask about a corpus.
    # llm, temperature, base_url, api_key, log
    ask_parser = parsers.add_parser(
        "ask",
        help=dedent(
            """\
            Ask questions about a corpus. Vectorstore will update when `ask` is
            invoked. Files that are in corpus, but not indexed will be indexed.
            Files that are indexed but not in corpus will be removed. Files
            that were modified after indexed will be re-indexed. 
            """
        ),
        formatter_class=BragCliFormatter,
    )

    add_llm_arg_parsers(ask_parser)
    add_corpus_dir(ask_parser)
    add_emb_arg_parsers(ask_parser)
    add_chunksize(ask_parser)
    add_chunk_overlap(ask_parser)
    add_batchsize(ask_parser)
    add_db_dir(ask_parser)
    add_db_name(ask_parser)
    add_db_host(ask_parser)
    add_db_port(ask_parser)
    add_num_retrieved_docs(ask_parser)
    add_no_cache_db(ask_parser)
    add_ssl_verify(ask_parser)
    add_log(ask_parser)

    ask_parser.add_argument(
        "--rag-type",
        choices=["brag", "trag"],
        help=dedent(
            """\
            * brag (basic rag, which does not use tool calling)
            * trag (tool-calling rag)
            """
        ),
        default="brag",
    )
    ask_parser.add_argument(
        "-p",
        "--port",
        help="port to serve web app. If not supplied, runs only in terminal.",
        type=int,
    )

    ask_parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help=dedent(
            """\
            Whether or not to show rag sources. If present, sources will be
            shown in command line. Sources can still be recovered via !cite.
            However, sources are always presented in the web app.
            """
        ),
        default=False,
    )

    ask_parser.add_argument(
        "--system-prompt",
        type=str,
        help=dedent(
            """\
            System prompt to use. Overriden if --system-prompt-path is supplied.
              * basic: answers based only on context.
              * mindful: answers based on context and llm's knowledge
            """
        ),
        choices=["basic", "mindful"],
        default="basic",
    )

    ask_parser.add_argument(
        "--system-prompt-path",
        type=str,
        help=dedent(
            """\
            Optional path to system prompt. The system prompt should be in a
            text file. If supplied, the --system-prompt flag will have no
            effect. If not supplied, then --system-prompt is applied.

            Example usage:
                brag ask --system-prompt-path=prompt.txt
            """
        ),
    )

    # Search within a corpus.
    search_parser = parsers.add_parser(
        "search",
        help="search within a corpus.",
        formatter_class=BragCliFormatter,
    )
    add_corpus_dir(search_parser)
    add_emb_arg_parsers(search_parser)
    add_chunksize(search_parser)
    add_chunk_overlap(search_parser)
    add_batchsize(search_parser)
    add_db_name(search_parser)
    add_db_dir(search_parser)
    add_db_host(search_parser)
    add_db_port(search_parser)
    add_num_retrieved_docs(search_parser)
    add_no_cache_db(search_parser)
    add_ssl_verify(search_parser)
    add_log(search_parser)

    # Chat.
    chat_parser = parsers.add_parser(
        "chat",
        help="chat with a bot.",
        formatter_class=BragCliFormatter,
    )
    add_llm_arg_parsers(chat_parser)
    add_ssl_verify(chat_parser)
    add_log(chat_parser)

    return parser.parse_args(), parser



"""Basic RAG (brag)."""

# NOTE: Putting the import statements near usage decreases time taken to
# generate help menus.

import logging
from pathlib import Path

from brag.parse_args import __version__, parse_args


def main():
    """Cli entry point."""
    args, parser = parse_args()

    if args.version:
        print(f"brag version {__version__}")
        exit()

    if args.command is None:
        parser.print_help()
        exit()

    # Set logging.
    numeric_level = getattr(logging, args.log.upper(), None)
    logging.basicConfig(
        level=numeric_level, format="[BRAG-%(levelname)s] %(message)s"
    )
    logging.debug(args)

    # from brag.util import make_emb, make_llm
    import litellm
    from langchain_litellm import ChatLiteLLM

    from brag.emb import LiteLLMEmbeddings

    match args.command:
        case "index":
            from brag.db import Database

            if args.no_ssl_verify:
                litellm.ssl_verify = False

            embedder = LiteLLMEmbeddings(
                model=args.emb,
                api_base=args.emb_base_url,
                api_key=args.emb_base_key,
            )
            Database(
                corpus_dir=Path(args.corpus_dir),
                embedder=embedder,
                index_doc_batch_size=args.batchsize,
                chunk_overlap=args.chunk_overlap,
                chunk_size=args.chunk_size,
                cache_db=True,
                db_dir=args.db_dir,
                db_name=args.db_name,
                db_host=args.db_host,
                db_port=args.db_port,
            )

        case "update-db":
            logging.warning("update-db (WIP)")

        case "ask":
            from brag.db import Database

            if args.no_ssl_verify:
                litellm.ssl_verify = False

            logging.info(args)
            llm = ChatLiteLLM(
                model=args.llm,
                api_key=args.api_key,
                api_base=args.base_url,
                temperature=args.temperature,
            )
            logging.info(f"Embedder: {args.emb}, base url: {args.emb_base_url}")

            embedder = LiteLLMEmbeddings(
                model=args.emb,
                api_base=args.emb_base_url,
                api_key=args.emb_api_key,
            )

            db = Database(
                corpus_dir=Path(args.corpus_dir),
                embedder=embedder,
                index_doc_batch_size=args.batchsize,
                num_retrieved_docs=args.num_retrieved_docs,
                chunk_overlap=args.chunk_overlap,
                chunk_size=args.chunk_size,
                cache_db=not args.no_cache_db,
                db_dir=args.db_dir,
                db_name=args.db_name,
                db_host=args.db_host,
                db_port=args.db_port,
                # Always show sources for web app.
                terminal=args.port is None,
            )
            match args.rag_type:
                case "brag":
                    from brag.rags import Brag

                    ChosenRag = Brag
                case "trag":
                    from brag.rags import Trag

                    ChosenRag = Trag
                case _:
                    raise ValueError(f"{args.rag_type} is not a valid option!")

            match args.system_prompt_path, args.system_prompt:
                case None, "basic":
                    from brag.rags.abstract import RAG_SYSTEM_PROMPT

                    system_prompt = RAG_SYSTEM_PROMPT
                case None, "mindful":
                    from brag.rags.abstract import MINDFUL_RAG_SYSTEM_PROMPT

                    system_prompt = MINDFUL_RAG_SYSTEM_PROMPT
                case None, _:
                    raise ValueError(
                        f"{args.system_prompt} is an invalid prompt."
                    )
                case _:  # system_prompt_path is supplied.
                    logging.debug(
                        "system_prompt_path is supplied and is:"
                        f"{args.system_prompt_path}"
                    )
                    with open(args.system_prompt_path, "r") as file:
                        system_prompt = file.read()
                    logging.debug(system_prompt)

            rag = ChosenRag(
                llm,
                db,
                thread_id=None,
                system_prompt=system_prompt,
                # Always show sources for web app.
                verbose=args.verbose or args.port is not None,
            )
            if args.port:
                from brag.ui import serve

                serve(rag, args)
            else:
                from brag.repl import AskREPL

                AskREPL().run(
                    "Ask questions about your corpus!",
                    rag.print_ask,
                )

        case "chat":
            from brag.chat import Chatbot
            from brag.repl import BragREPL

            if args.no_ssl_verify:
                litellm.ssl_verify = False

            logging.info(args)
            llm = ChatLiteLLM(
                model=args.llm,
                api_base=args.base_url,
                api_key=args.api_key,
            )
            chatbot = Chatbot(llm)
            BragREPL().run(
                f"Begin chatting with {args.llm} ", chatbot.print_stream
            )

        case "search":
            from brag.db import Database
            from brag.repl import BragREPL

            if args.no_ssl_verify:
                litellm.ssl_verify = False

            embedder = LiteLLMEmbeddings(
                model=args.emb,
                api_base=args.emb_base_url,
                api_key=args.emb_api_key,
            )

            db = Database(
                corpus_dir=Path(args.corpus_dir),
                embedder=embedder,
                index_doc_batch_size=args.batchsize,
                num_retrieved_docs=args.num_retrieved_docs,
                chunk_overlap=args.chunk_overlap,
                chunk_size=args.chunk_size,
                cache_db=not args.no_cache_db,
                db_dir=args.db_dir,
                db_name=args.db_name,
                db_host=args.db_host,
                db_port=args.db_port,
                terminal=True,
            )

            def print_results(query: str):
                print(db.retrieve(query, None))

            BragREPL().run("Begin searching your corpus!", print_results)

        case "rm-index":
            from chromadb import HttpClient, PersistentClient
            from chromadb.config import Settings

            # Recursively remove db dir.
            db_dir = Path(args.db_dir)
            local_files_to_remove = [
                db_dir / "corpus_mtimes.sqlite3",
                db_dir / "index-info.txt",
            ]
            for file in local_files_to_remove:
                try:
                    file.unlink()
                    logging.info(f"Successfully deleted '{file}'.")
                except FileNotFoundError:
                    logging.info(f"{file} was not found.")

            # Remove collection if needed.
            settings = Settings(allow_reset=True)
            if (port := args.db_port) is None:
                client = PersistentClient(path=str(db_dir), settings=settings)
            else:
                host = args.db_host or "localhost"
                client = HttpClient(host=host, port=port, settings=settings)

            try:
                # FIXME: the actual directories aren't removed, though the
                # collections are no longer associated. This could lead to large
                # disk space consumption.
                print(
                    "Number of collections (before):",
                    client.count_collections(),
                )

                client.delete_collection(name=args.db_name)
                if args.reset:
                    print("HERE")
                    client.reset()

                print(
                    "Number of collections (after):",
                    client.count_collections(),
                )

                logging.info(
                    f"Collection '{args.db_name}' deleted successfully."
                )
            except ValueError:
                logging.info(
                    f"Collection '{args.db_name}' was not found at {host}:{port}."
                )
